@@ -8,36 +8,68 @@ or
     restviewhttp [options] directory
 or
     restviewhttp [options] -e "command"
+or
+    restviewhttp [options] --long-description
 or
     restviewhttp --help
 
-Needs docutils and a web browser.  Will syntax highlight if you have pygments
-installed.
+Needs docutils and a web browser. Will syntax-highlight code or doctest blocks
+if you have pygments installed.
 """
+from __future__ import print_function
 
 import os
 import re
 import sys
+import time
 import socket
 import optparse
 import threading
 import webbrowser
-import BaseHTTPServer
-import cgi
-import urllib
+from contextlib import closing
+
+try:
+    import BaseHTTPServer
+except ImportError:
+    import http.server as BaseHTTPServer
+
+try:
+    import SocketServer
+except ImportError:
+    import socketserver as SocketServer
+
+try:
+    from html import escape
+except ImportError:
+    from cgi import escape
+
+try:
+    from urllib import unquote
+except ImportError:
+    from urllib.parse import unquote
+
+try:
+    from urlparse import parse_qs
+except ImportError:
+    from urllib.parse import parse_qs
 
 import docutils.core
 import docutils.writers.html4css1
 
 try:
     import pygments
-    import pygments.lexers
-    import pygments.formatters
+    from pygments import lexers, formatters
 except ImportError:
     pygments = None
 
 
-__version__ = "1.2.2"
+try:
+    unicode
+except NameError:
+    unicode = str
+
+
+__version__ = "2.0"
 
 
 class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
@@ -51,10 +83,13 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
             self.wfile.write(content)
 
     def do_HEAD(self):
-        content = self.do_GET_or_HEAD()
+        self.do_GET_or_HEAD()
 
     def do_GET_or_HEAD(self):
-        self.path = urllib.unquote(self.path)
+        self.path = unquote(self.path)
+        if '..' in self.path:
+            self.send_error(400, "Bad request") # no hacking!
+            return
         root = self.server.renderer.root
         command = self.server.renderer.command
         if self.path == '/':
@@ -67,20 +102,49 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
                     return self.handle_rest_file(root)
             else:
                 return self.handle_list(root)
-        elif '..' in self.path:
-            self.send_error(404, "File not found") # no hacking!
+        elif self.path.startswith('/polling?'):
+            query = parse_qs(self.path.partition('?')[-1])
+            pathname = query['pathname'][0]
+            if pathname == '/' and isinstance(root, str):
+                pathname = root
+            else:
+                pathname = self.translate_path(pathname)
+            old_mtime = int(query['mtime'][0])
+            return self.handle_polling(pathname, old_mtime)
+        elif self.path.endswith('.gif'):
+            return self.handle_image(self.translate_path(), 'image/gif')
         elif self.path.endswith('.png'):
             return self.handle_image(self.translate_path(), 'image/png')
-        elif self.path.endswith('.jpg'):
+        elif self.path.endswith('.jpg') or self.path.endswith('.jpeg'):
             return self.handle_image(self.translate_path(), 'image/jpeg')
         elif self.path.endswith('.txt') or self.path.endswith('.rst'):
             return self.handle_rest_file(self.translate_path())
         else:
-            self.send_error(404, "File not found: %s" % self.path)
+            self.send_error(501, "File type not supported: %s" % self.path)
 
-    def translate_path(self):
+    def handle_polling(self, path, old_mtime):
+        # TODO: use inotify if available
+        while True:
+            mtime = int(os.stat(path).st_mtime)
+            # we lose precision by using int(), but I'm nervous of
+            # round-tripping floating point numbers through HTML and
+            # comparing them for equality
+            if mtime != old_mtime:
+                try:
+                    self.send_response(200)
+                    self.send_header("Cache-Control", "no-cache, no-store, max-age=0")
+                    self.end_headers()
+                except Exception as e:
+                    self.log_error('%s (client closed "%s" before acknowledgement)', e, self.path)
+                finally:
+                    return
+            time.sleep(0.2)
+
+    def translate_path(self, path=None):
         root = self.server.renderer.root
-        path = self.path.lstrip('/')
+        if path is None:
+            path = self.path
+        path = path.lstrip('/')
         if not isinstance(root, str):
             idx, path = path.split('/', 1)
             root = root[int(idx)]
@@ -90,9 +154,10 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
 
     def handle_image(self, filename, ctype):
         try:
-            data = file(filename, 'rb').read()
+            with open(filename, 'rb') as f:
+                data = f.read()
         except IOError:
-            self.send_error(404, "File not found")
+            self.send_error(404, "File not found: %s" % self.path)
         else:
             self.send_response(200)
             self.send_header("Content-Type", ctype)
@@ -102,33 +167,29 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
 
     def handle_rest_file(self, filename):
         try:
-            f = open(filename)
-            try:
-                return self.handle_rest_data(f.read())
-            finally:
-                f.close()
-        except IOError, e:
+            with open(filename) as f:
+                mtime = os.fstat(f.fileno()).st_mtime
+                return self.handle_rest_data(f.read(), mtime=mtime)
+        except IOError as e:
             self.log_error("%s", e)
-            self.send_error(404, "File not found")
+            self.send_error(404, "File not found: %s" % self.path)
 
     def handle_command(self, command):
         try:
-            f = os.popen(command)
-            try:
+            with closing(os.popen(command)) as f:
                 return self.handle_rest_data(f.read())
-            finally:
-                f.close()
-        except OSError, e:
-            self.log_error("%s" % e)
-            self.send_error(404, "Command execution failed")
+        except OSError as e:
+            self.log_error("%s", e)
+            self.send_error(500, "Command execution failed")
 
-    def handle_rest_data(self, data):
-        html = self.server.renderer.rest_to_html(data)
+    def handle_rest_data(self, data, mtime=None):
+        html = self.server.renderer.rest_to_html(data, mtime=mtime)
         if isinstance(html, unicode):
             html = html.encode('UTF-8')
         self.send_response(200)
         self.send_header("Content-Type", "text/html; charset=UTF-8")
         self.send_header("Content-Length", str(len(html)))
+        self.send_header("Cache-Control", "no-cache, no-store, max-age=0")
         self.end_headers()
         return html
 
@@ -138,17 +199,18 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
         files = []
         for dirpath, dirnames, filenames in os.walk(dirname):
             dirnames[:] = [dn for dn in dirnames
-                           if dn != '.svn' and not dn.endswith('.egg-info')]
+                           if not dn.startswith('.')
+                           and not dn.endswith('.egg-info')]
             for fn in filenames:
                 if fn.endswith('.txt') or fn.endswith('.rst'):
                     prefix = dirpath[len(dirname):]
                     files.append(os.path.join(prefix, fn))
-        files.sort()
+        files.sort(key=str.lower)
         return files
 
     def handle_dir(self, dirname):
         files = [(fn, fn) for fn in self.collect_files(dirname)]
-        html = self.render_dir_listing('RST files in %s' % dirname, files)
+        html = self.render_dir_listing("RST files in %s" % os.path.abspath(dirname), files)
         if isinstance(html, unicode):
             html = html.encode('UTF-8')
         self.send_response(200)
@@ -163,11 +225,11 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
             if os.path.isdir(fn):
                 files.extend([(os.path.join(str(idx), f),
                                os.path.join(fn, f))
-                               for f in self.collect_files(fn)])
+                              for f in self.collect_files(fn)])
             else:
                 files.append((os.path.join(str(idx), os.path.basename(fn)),
                               fn))
-        html = self.render_dir_listing('RST files', files)
+        html = self.render_dir_listing("RST files", files)
         if isinstance(html, unicode):
             html = html.encode('UTF-8')
         self.send_response(200)
@@ -177,32 +239,61 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
         return html
 
     def render_dir_listing(self, title, files):
-        files = ''.join([FILE_TEMPLATE.replace('$href', cgi.escape(href))
-                                      .replace('$file', cgi.escape(fn))
+        files = ''.join([FILE_TEMPLATE.replace('$href', escape(href))
+                                      .replace('$file', escape(fn))
                          for href, fn in files])
-        return (DIR_TEMPLATE.replace('$title', cgi.escape(title))
+        return (DIR_TEMPLATE.replace('$title', escape(title))
                             .replace('$files', files))
 
 
 DIR_TEMPLATE = """\
+<!DOCTYPE html>
 <html>
-<head><title>$title</title></head>
+<head>
+<title>$title</title>
+</head>
 <body>
 <h1>$title</h1>
 <ul>
-  $files
-</ul>
+$files</ul>
 </body>
 </html>
 """
 
 FILE_TEMPLATE = """\
-  <li><a href="$href">$file</a></li>\
+  <li><a href="$href">$file</a></li>
+"""
+
+AJAX_STR = """
+<script type="text/javascript">
+var xmlHttp = null;
+window.onload = function () {
+    setTimeout(function () {
+        if (window.XMLHttpRequest) {
+            xmlHttp = new XMLHttpRequest();
+        } else if (window.ActiveXObject) {
+            xmlHttp = new ActiveXObject('Microsoft.XMLHTTP');
+        }
+        xmlHttp.onreadystatechange = function () {
+            if (xmlHttp.readyState == 4 && xmlHttp.status == '200') {
+                window.location.reload(true);
+            }
+        }
+        xmlHttp.open('HEAD', '/polling?pathname=' + location.pathname + '&mtime=%d', true);
+        xmlHttp.send(null);
+    }, 0);
+}
+window.onbeforeunload = function () {
+    xmlHttp.abort();
+}
+</script>
 """
 
 ERROR_TEMPLATE = """\
+<!DOCTYPE html>
 <html>
-<head><title>$title</title></head>
+<head>
+<title>$title</title>
 <style type="text/css">
 pre.error {
     border-left: 3px double red;
@@ -213,6 +304,7 @@ pre.error {
     color: red;
 }
 </style>
+</head>
 <body>
 <h1>$title</h1>
 <pre class="error">
@@ -226,10 +318,14 @@ $source
 """
 
 
+class ThreadingHTTPServer(SocketServer.ThreadingMixIn, BaseHTTPServer.HTTPServer):
+    daemon_threads = True
+
+
 class RestViewer(object):
     """Web server that renders ReStructuredText on the fly."""
 
-    server_class = BaseHTTPServer.HTTPServer
+    server_class = ThreadingHTTPServer
     handler_class = MyRequestHandler
 
     local_address = ('localhost', 0)
@@ -239,6 +335,8 @@ class RestViewer(object):
     css_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             'default.css')
 
+    strict = None
+
     def __init__(self, root, command=None):
         self.root = root
         self.command = command
@@ -259,34 +357,53 @@ class RestViewer(object):
         """
         self.server.serve_forever()
 
-    def rest_to_html(self, rest_input):
+    def close(self):
+        self.server.server_close()
+
+    def rest_to_html(self, rest_input, settings=None, mtime=None):
         """Render ReStructuredText."""
         writer = docutils.writers.html4css1.Writer()
         if pygments is not None:
             writer.translator_class = SyntaxHighlightingHTMLTranslator
         if self.css_url:
-            settings_overrides={'stylesheet': self.css_url,
-                                'stylesheet_path': None,
-                                'embed_stylesheet': False}
+            settings_overrides = {'stylesheet': self.css_url,
+                                  'stylesheet_path': None,
+                                  'embed_stylesheet': False}
         elif self.css_path:
-            settings_overrides={'stylesheet': self.css_path,
-                                'stylesheet_path': None,
-                                'embed_stylesheet': True}
+            settings_overrides = {'stylesheet': self.css_path,
+                                  'stylesheet_path': None,
+                                  'embed_stylesheet': True}
         else:
             settings_overrides = {}
+        settings_overrides['syntax_highlight'] = 'short'
+        if self.strict:
+            settings_overrides['halt_level'] = 1
+
+        if settings:
+            settings_overrides.update(settings)
+
         try:
             docutils.core.publish_string(rest_input, writer=writer,
                                          settings_overrides=settings_overrides)
-        except Exception, e:
+        except Exception as e:
             return self.render_exception(e.__class__.__name__, str(e),
                                          rest_input)
-        return writer.output
+        return self.get_markup(writer.output, mtime=mtime)
 
     def render_exception(self, title, error, source):
-        html = (ERROR_TEMPLATE.replace('$title', cgi.escape(title))
-                              .replace('$error', cgi.escape(error))
-                              .replace('$source', cgi.escape(source)))
-        return html
+        html = (ERROR_TEMPLATE.replace('$title', escape(title))
+                              .replace('$error', escape(error))
+                              .replace('$source', escape(source)))
+        return self.get_markup(html)
+
+    def get_markup(self, markup, mtime=None):
+        if self.command is not None:
+            return markup.replace('</title>',
+                                  ' -e "' + escape(self.command) + '"</title>')
+        elif mtime is not None:
+            return markup.replace('</body>', (AJAX_STR % mtime) + '</body>')
+        else:
+            return markup
 
 
 class SyntaxHighlightingHTMLTranslator(docutils.writers.html4css1.HTMLTranslator):
@@ -294,6 +411,11 @@ class SyntaxHighlightingHTMLTranslator(docutils.writers.html4css1.HTMLTranslator
     in_doctest = False
     in_text = False
     in_reference = False
+    formatter_styles = formatters.HtmlFormatter(style='colorful').get_style_defs('pre')
+
+    def __init__(self, document):
+        docutils.writers.html4css1.HTMLTranslator.__init__(self, document)
+        self.body_prefix[:0] = ['<style type="text/css">\n', self.formatter_styles, '\n</style>\n']
 
     def visit_doctest_block(self, node):
         docutils.writers.html4css1.HTMLTranslator.visit_doctest_block(self, node)
@@ -306,12 +428,8 @@ class SyntaxHighlightingHTMLTranslator(docutils.writers.html4css1.HTMLTranslator
     def visit_Text(self, node):
         if self.in_doctest:
             text = node.astext()
-            lexer = pygments.lexers.PythonConsoleLexer()
-            # noclasses forces inline styles, which is suboptimal
-            # figure out a way to include formatter.get_style_defs() into
-            # our CSS
-            formatter = pygments.formatters.HtmlFormatter(nowrap=True,
-                                                          noclasses=True)
+            lexer = lexers.PythonConsoleLexer()
+            formatter = formatters.HtmlFormatter(nowrap=True)
             self.body.append(pygments.highlight(text, lexer, formatter))
         else:
             text = node.astext()
@@ -324,8 +442,10 @@ class SyntaxHighlightingHTMLTranslator(docutils.writers.html4css1.HTMLTranslator
 
     def visit_literal(self, node):
         self.in_text = True
-        docutils.writers.html4css1.HTMLTranslator.visit_literal(self, node)
-        self.in_text = False
+        try:
+            docutils.writers.html4css1.HTMLTranslator.visit_literal(self, node)
+        finally:
+            self.in_text = False
 
     def visit_reference(self, node):
         self.in_reference = True
@@ -348,11 +468,15 @@ class SyntaxHighlightingHTMLTranslator(docutils.writers.html4css1.HTMLTranslator
             >>> link_local_files = SyntaxHighlightingHTMLTranslator.link_local_files
             >>> link_local_files('e.g. see README.txt for more info')
             'e.g. see <a href="README.txt">README.txt</a> for more info'
+            >>> link_local_files('e.g. see docs/HACKING.rst for more info')
+            'e.g. see <a href="docs/HACKING.rst">docs/HACKING.rst</a> for more info'
+            >>> link_local_files('what about http://example.com/README.txt ?')
+            'what about http://example.com/README.txt ?'
 
         """
         # jwz was right...
-        text = re.sub("([-_a-zA-Z0-9]+[.]txt)", r'<a href="\1">\1</a>', text)
-        return text
+        return re.sub(r"(^|\s)([-_a-zA-Z0-9/]+[.](txt|rst))",
+                      r'\1<a href="\2">\2</a>', text)
 
 
 def parse_address(addr):
@@ -368,11 +492,11 @@ def parse_address(addr):
         ('', 1234)
 
         >>> try: parse_address('notanumber')
-        ... except ValueError, e: print e
+        ... except ValueError as e: print(e)
         Invalid address: notanumber
 
         >>> try: parse_address('la:la:la')
-        ... except ValueError, e: print e
+        ... except ValueError as e: print(e)
         Invalid address: la:la:la
 
     """
@@ -401,18 +525,29 @@ def get_host_name(listen_on):
         ip_addr = socket.inet_aton(listen_on)
     except socket.error: # probably a hostname or ''
         ip_addr = None
-    if listen_on == '' or ip_addr == '\0\0\0\0':
+    if listen_on == '' or ip_addr == b'\0\0\0\0':
         return socket.gethostname()
     else:
         return listen_on
 
 
+def launch_browser(url):
+    """Launch the web browser for a given URL.
+
+    Does not block.
+    """
+    # Do it in the background as it may block
+    t = threading.Thread(target=webbrowser.open, args=(url,))
+    t.setDaemon(True)
+    t.start()
+
+
 def main():
     progname = os.path.basename(sys.argv[0])
     parser = optparse.OptionParser("%prog [options] filename-or-directory [...]",
                     description="Serve ReStructuredText files over HTTP.",
                     prog=progname)
-    parser.add_option('-l', '--listen',
+    parser.add_option('-l', '--listen', metavar='PORT',
                       help='listen on a given port (or interface:port,'
                            ' e.g. *:8080) [default: random port on localhost]',
                       default=None)
@@ -420,12 +555,19 @@ def main():
                       help='open a web browser [default: only if -l'
                            ' was not specified]',
                       action='store_true', default=None)
-    parser.add_option('-e', '--execute',
+    parser.add_option('-e', '--execute', metavar='COMMAND',
                       help='run a command to produce ReStructuredText',
                       default=None)
-    parser.add_option('--css',
+    parser.add_option('--long-description',
+                      help='run "python setup.py --long-description" to produce ReStructuredText',
+                      action='store_const', dest='execute',
+                      const='python setup.py --long-description')
+    parser.add_option('--css', metavar='URL-or-FILENAME',
                       help='use the specified stylesheet',
                       action='store', dest='css_path', default=None)
+    parser.add_option('--strict',
+                      help='halt at the slightest problem',
+                      action='store_true', default=False)
     opts, args = parser.parse_args(sys.argv[1:])
     if not args and not opts.execute:
         parser.error("at least one argument expected")
@@ -447,24 +589,26 @@ def main():
         else:
             server.css_path = opts.css_path
             server.css_url = None
+
+    server.strict = opts.strict
+
     if opts.listen:
         try:
             server.local_address = parse_address(opts.listen)
-        except ValueError, e:
-            sys.exit(str(e))
+        except ValueError as e:
+            parser.error(str(e))
     host = get_host_name(server.local_address[0])
     port = server.listen()
-    url = 'http://%s:%d/' % (host, port)
-    print "Listening on %s" % url
-    if opts.browser:
-        # launch the web browser in the background as it may block
-        t = threading.Thread(target=webbrowser.open, args=(url,))
-        t.setDaemon(True)
-        t.start()
     try:
+        url = 'http://%s:%d/' % (host, port)
+        print("Listening on %s" % url)
+        if opts.browser:
+            launch_browser(url)
         server.serve()
     except KeyboardInterrupt:
         pass
+    finally:
+        server.close()
 
 
 if __name__ == '__main__':
